--------------------------- MODULE digital_watch ---------------------------
(* Notes: Adaptation of the digital watch model presented by Harel (87)

To cite this model, please use:

@inproceedings{DBLP:conf/re/AbbassiBDS18,
  author    = {Ali Abbassi and
               Amin Bandali and
               Nancy A. Day and
               Jos{\'{e}} Serna},
  title     = {A Comparison of the Declarative Modelling Languages B,
                  Dash, and {TLA+}},
  booktitle = {8th {IEEE} International Model-Driven Requirements
                  Engineering Workshop, MoDRE@RE 2018, Banff, AB,
                  Canada, August 20, 2018},
  pages     = {11--20},
  year      = {2018},
  crossref  = {DBLP:conf/re/2018modre},
  url       = {https://doi.org/10.1109/MoDRE.2018.00008},
  doi       = {10.1109/MoDRE.2018.00008},
  timestamp = {Thu, 14 Feb 2019 15:07:29 +0100},
  biburl    = {https://dblp.org/rec/bib/conf/re/AbbassiBDS18},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}

@proceedings{DBLP:conf/re/2018modre,
  editor    = {Ana Moreira and
               Gunter Mussbacher and
               Jo{\~{a}}o Ara{\'{u}}jo and
               Pablo S{\'{a}}nchez},
  title     = {8th {IEEE} International Model-Driven Requirements
                  Engineering Workshop, MoDRE@RE 2018, Banff, AB,
                  Canada, August 20, 2018},
  publisher = {{IEEE} Computer Society},
  year      = {2018},
  url       = {http://ieeexplore.ieee.org/xpl/mostRecentIssue.jsp?punumber=8500723},
  isbn      = {978-1-5386-8406-1},
  timestamp = {Thu, 14 Feb 2019 15:07:29 +0100},
  biburl    = {https://dblp.org/rec/bib/conf/re/2018modre},
  bibsource = {dblp computer science bibliography, https://dblp.org}
} *)

EXTENDS Integers, TLC, FiniteSets, Sequences

VARIABLES
    light, status,
    waited_2_min, waited_2_sec,
    pressed

STATUS == { "Time", "Date", "Wait", "Update"
          , "Alarm1", "Alarm2", "Chime", "StopWatch"
          , "Alarms_Beep" (* doesn't seem to be used right now *)  }

KEYS == { "a", "b", "c", "d" }

vars == << light, status, waited_2_min, waited_2_sec, pressed >>

vars_but_light == << status, waited_2_min, waited_2_sec, pressed >>

vars_but_status == << light, waited_2_min, waited_2_sec, pressed >>


\* Helper predicate for range of a function
Range(f) == {f[x]: x \in DOMAIN f}


TypeOK ==  \* Typing invariant
    /\ light \in BOOLEAN  \* FALSE: Off, TRUE: On
    /\ status \in STATUS
    /\ waited_2_min \in BOOLEAN /\ waited_2_sec \in BOOLEAN
    /\ pressed \in [KEYS -> BOOLEAN]

Init ==  \* Initial state
    /\ light = FALSE  \* initially light is off
    /\ status = "Time"  \* initially display shows time
    /\ waited_2_min = FALSE /\ waited_2_sec = FALSE
    /\ pressed = [k \in KEYS |-> FALSE]


\* <Light>
light_off_light_on ==
    /\ pressed["b"]
    /\ light' = TRUE
    /\ UNCHANGED vars_but_light

light_on_light_off ==
    /\ ~pressed["b"]
    /\ light' = FALSE
    /\ UNCHANGED vars_but_light
\* </Light>


\* <Time>
time_show_date ==
    /\ status = "Time"
    /\ pressed["d"]
    /\ status' = "Date"
    /\ UNCHANGED vars_but_status

time_try_update ==
    /\ status = "Time"
    /\ pressed["c"]
    /\ status' = "Wait"
    /\ UNCHANGED vars_but_status

time_go2alarm1 ==
    /\ status = "Time"
    /\ pressed["a"]
    /\ status' = "Alarm1"
    /\ UNCHANGED vars_but_status
\* </Time>


\* <Date>
date_show_time ==
    /\ status = "Date"
    /\ pressed["d"]
    /\ status' = "Time"
    /\ UNCHANGED vars_but_status

date_return_to_time ==
    /\ status = "Date" 
    /\ waited_2_min
    /\ status' = "Time"
    /\ UNCHANGED vars_but_status
\* </Date>


\* <Wait>
wait_show_time ==
    /\ status = "Wait"
    /\ ~pressed["c"]
    /\ status' = "Time"
    /\ UNCHANGED vars_but_status

wait_show_update == 
    /\ status = "Wait"
    /\ waited_2_sec
    /\ status' = "Update"
    /\ UNCHANGED vars_but_status
\* </Wait>


\* <Update>
update_show_time ==
    /\ status = "Update"
    /\ pressed["b"]
    /\ status' = "Time"
    /\ UNCHANGED vars_but_status
\* </Update>


\* <Alarm1>
alarm1_go2alarm2 ==
    /\ status = "Alarm1"
    /\ pressed["a"]
    /\ status' = "Alarm2"
    /\ UNCHANGED vars_but_status
\* </Alarm1>


\* <Alarm2>
alarm2_go2chime ==
    /\ status = "Alarm2"
    /\ pressed["a"]
    /\ status' = "Chime"
    /\ UNCHANGED vars_but_status
\* </Alarm2>


\* <Chime>
chime_go2Stopwatch ==
    /\ status = "Chime"
    /\ pressed["a"]
    /\ status' = "StopWatch"
    /\ UNCHANGED vars_but_status
\* </Chime>


\* <StopWatch>
Stopwatch_go2time ==
    /\ status = "StopWatch"
    /\ pressed["a"]
    /\ status' = "Time"
    /\ UNCHANGED vars_but_status
\* </StopWatch>


\* <Alarms_Beep>
\* </Alarms_Beep>


\* <Helpers>

\* Key presses
PressKey(k) ==
    /\ ~pressed[k] /\ pressed' = [pressed EXCEPT ![k] = TRUE]
    /\ UNCHANGED << light, status, waited_2_min, waited_2_sec >>
ReleaseKey(k) ==
    /\ pressed[k] /\ pressed' = [pressed EXCEPT ![k] = FALSE]
    /\ UNCHANGED << light, status, waited_2_min, waited_2_sec >>

\* Waits
waited_2_min_t == ~waited_2_min /\ waited_2_min' = TRUE
    /\ UNCHANGED << light, status, waited_2_sec, pressed >>
waited_2_min_f == waited_2_min /\ waited_2_min' = FALSE
    /\ UNCHANGED << light, status, waited_2_sec, pressed >>

waited_2_sec_t == ~waited_2_sec /\ waited_2_sec' = TRUE
    /\ UNCHANGED << light, status, waited_2_min, pressed >>
waited_2_sec_f == waited_2_sec /\ waited_2_sec' = FALSE
    /\ UNCHANGED << light, status, waited_2_min, pressed >>
\* </Helpers>


\* <Temporal properties>

(***************************************************************************)
(* I believe the original eventually_time property from Dash expressed in  *)
(* CTL says that "on a press_a, it's possible that in the future the       *)
(* display will display the time".  However, since TLA+'s temporal logic   *)
(* is LTL-based and not CTL, we can't easily express possibility           *)
(* properties.  So, instead, we'll state that "on a press_a, in the future *)
(* the display will display the time".                                     *)
(***************************************************************************)

EventuallyTime ==
    [] (pressed["a"] => <> (status = "Time"))
\* Note: the above property does NOT hold with weak or strong fairness on
\* all the actions on vars

\* </Temporal properties>


\* Spec

Next ==
    \/ light_off_light_on \/ light_on_light_off
    \/ time_show_date \/ time_try_update \/ time_go2alarm1
    \/ date_show_time \/ date_return_to_time
    \/ wait_show_time \/ wait_show_update
    \/ update_show_time
    \/ alarm1_go2alarm2
    \/ alarm2_go2chime
    \/ chime_go2Stopwatch
    \/ Stopwatch_go2time
    \/ \E k \in KEYS: PressKey(k)
    \/ \E k \in KEYS: ReleaseKey(k)
    \/ waited_2_min_t \/ waited_2_min_f
    \/ waited_2_sec_t \/ waited_2_sec_f

Live == WF_vars(Next)

Spec == Init /\ [][Next]_vars /\ Live  (************************************)
                                       (* every transition either          *)
                                       (* satisfies the action formula     *)
                                       (* `Next' or leaves the expression  *)
                                       (* `vars' unchanged.  In            *)
                                       (* particular, this admits          *)
                                       (* "stuttering transitions" that do *)
                                       (* not affect `vars'.  That is to   *)
                                       (* say, [][Next]_vars == [](Next \/ *)
                                       (* (vars' = vars))                  *)
                                       (************************************) 

=============================================================================
\* Modification History
\* Last modified Tue Jul 17 14:04:48 EDT 2018 by amin
\* Created Tue May 29 18:29:07 EDT 2018 by amin
