------------------------------ MODULE ehealth ------------------------------
(* To cite this model, please use:

@techreport{ostroff2017,
  author      = {Jonathan S. Ostroff},
  title       = {Validating Software via Abstract State Specifications},
  year        = {2017},
  number      = {EECS-2017-02},
  institution = {York University},
  url         = {http://www.eecs.yorku.ca/research/techreports/2017/?abstract=EECS-2017-02}
}

  This model was created by Prof. J. Ostroff at York U (Canada) and
  used for teaching EECS 4312, Software Engineering Requirements.
  The model was used with explicit permission from Prof. Ostroff.

*)

EXTENDS Naturals, Sequences, TLC

CONSTANTS
    M \* set of all medications in the universe
  , P \* set of all patients in the universe

VARIABLES
    medications \* set of medications registered in the system
  , patients    \* set of patients registered in the system
  , prescriptions
  , interactions

vars == <<medications, patients, prescriptions, interactions>>

TypeOK ==
    /\ patients \subseteq P
    /\ medications \subseteq M
    /\ prescriptions \subseteq patients \X medications
    /\ interactions  \subseteq medications \X medications \* dangerous interactions

\* Invariants

Symmetry == \A m1, m2 \in medications:
    <<m1, m2>> \in interactions <=>
    <<m2, m1>> \in interactions

\* A medication does not interact with itself
Irreflexive == \A m \in medications:
    <<m, m>> \notin interactions

\* All prescriptions are safe
\* i.e. do not have dangerous interactions
SafePrescriptions ==
    \A m1, m2 \in medications, p \in patients:
        <<m1, m2>> \in interactions =>
            ~(<<p, m1>> \in prescriptions
            /\ <<p, m2>> \in prescriptions)

\* Actions

add_patient(p) ==
    IF
        /\ p \in P
        /\ p \notin patients
    THEN
        /\ patients' = patients \union {p}
        /\ UNCHANGED <<medications, interactions, prescriptions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"add_patient", "e1">>, TRUE)

add_medication(m) ==
    IF
        /\ m \in M
        /\ m \notin medications
    THEN
        /\ medications' = medications \union {m}
        /\ UNCHANGED <<patients, interactions, prescriptions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"add_medication", "e2">>, TRUE)

add_interaction(m1, m2) ==
    IF
        /\ m1 \in medications
        /\ m2 \in medications
        /\ m1 /= m2
        /\ <<m1, m2>> \notin interactions
        /\ \A p \in patients: ~(<<p, m1>> \in prescriptions
                              /\ <<p, m2>> \in prescriptions)
    THEN
        /\ interactions' = interactions \union {<<m1, m2>>, <<m2, m1>>}
        /\ UNCHANGED <<medications, patients, prescriptions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"add_interaction", "e3">>, TRUE)

add_prescription(p, m) ==
    IF
        /\ p \in patients
        /\ m \in medications
        /\ <<p, m>> \notin prescriptions
        /\ \A x \in medications: <<p, x>> \in prescriptions =>
            <<x, m>> \notin interactions
    THEN
        /\ prescriptions' = prescriptions \union {<<p, m>>}
        /\ UNCHANGED <<medications, patients, interactions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"add_prescription", "e4">>, TRUE)

remove_interaction(m1, m2) ==
    IF
        /\ m1 \in medications
        /\ m2 \in medications
        /\ <<m1, m2>> \in interactions
    THEN
        /\ interactions' = interactions \ {<<m1, m2>>, <<m2, m1>>}
        /\ UNCHANGED <<medications, patients, prescriptions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"remove_interaction", "e5">>, TRUE)

remove_prescription(p, m) ==
    IF
        /\ p \in patients
        /\ m \in medications
        /\ <<p, m>> \in prescriptions
    THEN
        /\ prescriptions' = prescriptions \ {<<p, m>>}
        /\ UNCHANGED <<medications, patients, interactions>>
    ELSE
        /\ UNCHANGED vars
\*        /\ Print(<<"remove_prescription", "e6">>, TRUE)

\* Spec

Init ==
    /\ medications = {}
    /\ patients = {}
    /\ prescriptions = {}
    /\ interactions  = {}

Next ==
    \/ \E p \in P: add_patient(p)
    \/ \E m \in M: add_medication(m)
    \/ (\E m1, m2 \in M: add_interaction(m1, m2) \/ remove_interaction(m1, m2))
    \/ (\E m \in M, p \in P: add_prescription(p, m) \/ remove_prescription(p, m))

Spec == Init /\ [][Next]_vars

=============================================================================
\* Modification History
\* Last modified Fri Sep 27 23:30:46 EDT 2019 by bandali
\* Last modified Mon Jul 23 18:54:06 EDT 2018 by amin
\* Created Thu Nov 09 19:30:33 EST 2017 by amin
