------------------------------ MODULE library ------------------------------

EXTENDS Sequences, TLC, utils

CONSTANTS MemberID, BookID

VARIABLES members, books, loans, reservations

vars == << members, books, loans, reservations >>

maxNbLoans == Cardinality(BookID) - 1

TypeOK ==
    /\ members \subseteq MemberID
    /\ books \subseteq BookID
    /\ loans \in [books -> SUBSET members]
    /\ reservations \in [books -> Seq(members)]

Acquire(book) ==
    /\ book \in BookID
    /\ book \notin books
    /\ books' = books \union {book}
    /\ loans' = loans @@ book :> {}
    /\ reservations' = reservations @@ book :> <<>>
    /\ UNCHANGED members

Cancel(member, book) ==
    /\ member \in MemberID
    /\ book \in BookID
    /\ member \in members
    /\ book \in books
    /\ reservations[book] # <<>>
    /\ member \in Range(reservations[book])
    /\ reservations' =
        [reservations EXCEPT ![book] =
            LET s == reservations[book] IN
                LET i == Index(member, reservations[book]) IN
                    SubSeq(s, 1, i-1) \o SubSeq(s, i+1, Len(s))]
    /\ UNCHANGED << members, books, loans >>

Discard(book) ==
    /\ book \in BookID
    /\ book \in books
    /\ loans[book] = {}
    /\ reservations[book] = <<>>
    /\ books' = books \ {book}
    /\ loans' = DomSub({book}, loans)
    /\ reservations' = DomSub({book}, reservations)
    /\ UNCHANGED members

Join(member) ==
    /\ member \in MemberID
    /\ member \notin members
    /\ members' = members \union {member}
    /\ UNCHANGED << books, loans, reservations >>

Leave(member) ==
    /\ member \in MemberID
    /\ member \in members
    /\ {member} \notin Range(loans)
    /\ reservations # <<>> =>
        \A r \in Range(reservations): RanRes(r, {member}) = <<>>
    /\ members' = members \ {member}
    /\ UNCHANGED << books, loans, reservations >>

Lend(member, book) ==
    /\ member \in MemberID
    /\ book \in BookID
    /\ member \in members
    /\ book \in books
    /\ loans[book] = {}
    /\ reservations[book] = <<>>
    /\ Card(RanRes(loans, {{member}})) < maxNbLoans
    /\ loans' = [loans EXCEPT ![book] = {member}]
    /\ UNCHANGED << members, books, reservations >>

Renew(member, book) ==
    /\ member \in MemberID
    /\ book \in BookID
    /\ member \in members
    /\ book \in books
    /\ loans[book] = {member}
    /\ reservations[book] = <<>>
    /\ TRUE
    /\ UNCHANGED vars

Reserve(member, book) ==
    /\ member \in MemberID
    /\ book \in BookID
    /\ member \in members
    /\ book \in books
    /\ member \notin Range(reservations[book])
    /\ loans[book] # {member}
    /\ \/ loans[book] # {}
       \/ reservations[book] # <<>>
    /\ reservations' =
        [reservations EXCEPT ![book] = Append(reservations[book], member)]
    /\ UNCHANGED << members, books, loans >>

Return(book) ==
    /\ book \in BookID
    /\ book \in books
    /\ loans[book] # {}
    /\ loans' = [loans EXCEPT ![book] = {}]
    /\ UNCHANGED << members, books, reservations >>

Take(member, book) ==
    /\ member \in MemberID
    /\ book \in BookID
    /\ member \in members
    /\ book \in books
    /\ loans[book] = {}
    /\ Card(RanRes(loans, {{member}})) < maxNbLoans
    /\ Card(reservations[book]) # 0
    /\ Head(reservations[book]) = member
    /\ loans' = [loans EXCEPT ![book] = {member}]
    /\ reservations' =
        [reservations EXCEPT ![book] = Tail(reservations[book])]
    /\ UNCHANGED << members, books >>


Init ==
    /\ members = {}
    /\ books = {}
    /\ loans = <<>>
    /\ reservations = <<>>

Next ==
    \/ \E b \in BookID:
        \/ Acquire(b)
        \/ Discard(b)
        \/ Return(b)
    \/ \E m \in MemberID:
        \/ Join(m)
        \/ Leave(m)
    \/ \E m \in MemberID, b \in BookID:
        \/ Cancel(m, b)
        \/ Lend(m, b)
        \/ Renew(m, b)
        \/ Reserve(m, b)
        \/ Take(m, b)

Live == TRUE \* WF_vars(Next)

Spec == Init /\ [][Next]_vars /\ Live


ASSERT_LTL_2 ==
    \A b \in BookID:
        [](ENABLED Acquire(b) => b \notin books)

ASSERT_LTL_3 ==
    \A b \in BookID:
        [](ENABLED Discard(b) =>
            /\ loans[b] = {}
            /\ reservations[b] = <<>>)

ASSERT_LTL_4 ==
    \A m \in MemberID, b \in BookID:
        [](ENABLED Lend(m, b) => m \in members)

ASSERT_LTL_5 ==
    \A m \in MemberID, b \in BookID:
        [](ENABLED Reserve(m, b) =>
            /\ m \notin Range(reservations[b])
            /\ loans[b] # {m}
            /\ \/ loans[b] # {}
               \/ reservations[b] # <<>>)

ASSERT_LTL_6 ==
    \A m \in MemberID, b \in BookID:
        [](ENABLED Reserve(m, b) => loans[b] # {m})

ASSERT_LTL_7 ==
    \A m \in MemberID, b \in BookID:
        [](ENABLED Reserve(m, b) => m \notin Range(reservations[b]))

ASSERT_LTL_8 ==
    \A m \in MemberID, b \in BookID:
        [](\A b_ \in DOMAIN reservations:
            reservations[b_] # <<>> =>
                ~(ENABLED Lend(m, b) /\ b = b_))

ASSERT_LTL_9 ==
    \A m \in MemberID, b \in BookID:
        [](\A b_ \in DOMAIN reservations:
            reservations[b_] # <<>> =>
                ~(ENABLED Renew(m, b) /\ b = b_))

ASSERT_LTL_10 ==
    \A m \in MemberID, b \in BookID:
        [](ENABLED Take(m, b) =>
            /\ reservations[b] # <<>>
            /\ Head(reservations[b]) = m)

ASSERT_LTL_11 ==  \* not accurately represented, since TLA+ doesn't have W
    \A m \in MemberID, b \in BookID:
        /\ [](\A b_ \in DOMAIN loans:
            loans[b_] # {} => ~(ENABLED Take(m, b) /\ b = b_))
        /\ [](ENABLED Take(m, b) => loans[b] = {})

ASSERT_LTL_13 ==
    \A m \in MemberID:
        [](ENABLED Leave(m) =>
            /\ m \in members
            /\ {m} \notin Range(loans)
            /\ reservations # <<>> =>
                \A r \in Range(reservations): RanRes(r, {m}) = <<>>)

ASSERT_LTL_15 ==
    [](\A m \in members: Card(RanRes(loans, {{m}})) <= maxNbLoans)

=============================================================================
\* Modification History
\* Last modified Sat Jan 04 17:37:47 EST 2020 by bandali
\* Created Mon Dec 30 12:52:34 EST 2019 by bandali
