--------------------------- MODULE musical_chairs ---------------------------
(* Notes: Adaptation of the model presented by Nissanke (99)

To cite this model, please use:

@inproceedings{DBLP:conf/re/AbbassiBDS18,
  author    = {Ali Abbassi and
               Amin Bandali and
               Nancy A. Day and
               Jos{\'{e}} Serna},
  title     = {A Comparison of the Declarative Modelling Languages B,
                  Dash, and {TLA+}},
  booktitle = {8th {IEEE} International Model-Driven Requirements
                  Engineering Workshop, MoDRE@RE 2018, Banff, AB,
                  Canada, August 20, 2018},
  pages     = {11--20},
  year      = {2018},
  crossref  = {DBLP:conf/re/2018modre},
  url       = {https://doi.org/10.1109/MoDRE.2018.00008},
  doi       = {10.1109/MoDRE.2018.00008},
  timestamp = {Thu, 14 Feb 2019 15:07:29 +0100},
  biburl    = {https://dblp.org/rec/bib/conf/re/AbbassiBDS18},
  bibsource = {dblp computer science bibliography, https://dblp.org}
}

@proceedings{DBLP:conf/re/2018modre,
  editor    = {Ana Moreira and
               Gunter Mussbacher and
               Jo{\~{a}}o Ara{\'{u}}jo and
               Pablo S{\'{a}}nchez},
  title     = {8th {IEEE} International Model-Driven Requirements
                  Engineering Workshop, MoDRE@RE 2018, Banff, AB,
                  Canada, August 20, 2018},
  publisher = {{IEEE} Computer Society},
  year      = {2018},
  url       = {http://ieeexplore.ieee.org/xpl/mostRecentIssue.jsp?punumber=8500723},
  isbn      = {978-1-5386-8406-1},
  timestamp = {Thu, 14 Feb 2019 15:07:29 +0100},
  biburl    = {https://dblp.org/rec/bib/conf/re/2018modre},
  bibsource = {dblp computer science bibliography, https://dblp.org}
} *)

EXTENDS Integers, FiniteSets

CONSTANTS CHAIRS, PLAYERS

VARIABLES
    activePlayers, activeChairs,
    occupied, music_playing, state

vars == << activePlayers, activeChairs, occupied, music_playing, state >>

STATE_ == { "Start", "Walking", "Sitting", "Won" }

\* Helper predicate for range of a function
Range(f) == {f[x]: x \in DOMAIN f}


\* Typing invariant
TypeOK ==
    /\ activePlayers  \subseteq PLAYERS
    /\ activeChairs   \subseteq CHAIRS
    /\ occupied \in [activeChairs -> activePlayers] \union {<<>>}
    /\ music_playing \in BOOLEAN  \* whether music is playing
    /\ state \in STATE_

\* Initial state
Init ==
    /\ activePlayers = PLAYERS  \* force all activePlayers and
    /\ activeChairs = CHAIRS    \* activeChairs to be included
    /\ Cardinality(activePlayers) > 1
    /\ Cardinality(activePlayers) = Cardinality(activeChairs) + 1
    /\ occupied = <<>>  \* initially the empty function
    /\ music_playing \in BOOLEAN
    /\ state = "Start"

Walk ==
    /\ state = "Start"
    /\ music_playing
    /\ Cardinality(activePlayers) > 1
    /\ occupied' = <<>>
    /\ state' = "Walking"
    /\ UNCHANGED <<activeChairs, activePlayers, music_playing>>

Sit ==
    /\ state = "Walking"
    /\ ~music_playing
    /\ occupied' \in [activeChairs -> activePlayers]
    \* each chair maps to only one player
    /\ \A c \in activeChairs, p1, p2 \in activePlayers:
        occupied'[c] = p1 /\ occupied'[c] = p2 => p1 = p2
    \* each occupying player sits on one chair
    /\ \A p \in Range(occupied'), c1, c2 \in DOMAIN occupied':
        occupied'[c1] = p /\ occupied'[c2] = p => c1 = c2
    \* there's a player that didn't get to sit down
    \* /\ \E p \in activePlayers: p \notin Range(occupied')
    /\ state' = "Sitting"
    /\ UNCHANGED <<activeChairs, activePlayers, music_playing>>

EliminateLoser ==
    /\ state = "Sitting"
    /\ Cardinality(activePlayers) > 1
    /\ Cardinality(activePlayers) - Cardinality(Range(occupied)) = 1
    /\ activePlayers' = Range(occupied)
    /\ activeChairs' = activeChairs \ {CHOOSE c \in activeChairs : TRUE}
    /\ Cardinality(activeChairs') = Cardinality(activeChairs) - 1
    /\ occupied' = <<>>
    /\ state' = "Start"
    /\ UNCHANGED music_playing

Win ==
    /\ state = "Sitting"
    /\ Cardinality(activePlayers) = 1
    /\ state' = "Won"

ChangeMusicPlaying ==
    /\ music_playing' \in BOOLEAN
    /\ UNCHANGED <<activeChairs, activePlayers, state, occupied>>

\* Safety invariants

OneMorePlayerThanChairs ==
    Cardinality(activePlayers) = Cardinality(activeChairs) + 1

\* Temporal properties

ExistentialLiveness ==
    \E p \in PLAYERS: <> (activePlayers = {p})

FiniteLiveness == <> ENABLED Sit

InfiniteLiveness == <>[](Cardinality(activePlayers) = 1)


Next ==
    \/ Walk
    \/ Sit
    \/ EliminateLoser
    \/ Win
    \/ ChangeMusicPlaying

PlayActions ==
    \/ Walk
    \/ Sit
    \/ EliminateLoser
    \/ Win

Live == SF_vars(PlayActions) /\ WF_vars(ChangeMusicPlaying)
\*Live == TRUE  \* don't assume fairness

Spec == Init /\ [][Next]_vars /\ Live  \* every transition either satisfies the action
                                       \* formula `Next' or leaves the expression
                                       \* `vars' unchanged. In particular, this admits
                                       \* "stuttering transitions" that do not affect
                                       \* `vars'. That is to say,
                                       \* [][Next]_vars == [](Next \/ (vars' = vars)) 

=============================================================================
\* Modification History
\* Last modified Tue Jul 17 14:05:02 EDT 2018 by amin
\* Created Mon May 14 11:12:23 EDT 2018 by amin
