--------------------------- MODULE musical_chairs ---------------------------
EXTENDS Integers, FiniteSets

CONSTANTS CHAIRS, PLAYERS

VARIABLES
    activePlayers, activeChairs,
    occupied, music_playing, state

vars == << activePlayers, activeChairs, occupied, music_playing, state >>

STATE_ == { "Start", "Walking", "Sitting", "Won" }

\* Helper predicate for range of a function
Range(f) == {f[x]: x \in DOMAIN f}


\* Typing invariant
TypeOK ==
    /\ activePlayers  \subseteq PLAYERS
    /\ activeChairs   \subseteq CHAIRS
    /\ occupied \in [activeChairs -> activePlayers] \union {<<>>}
    /\ music_playing \in BOOLEAN  \* whether music is playing
    /\ state \in STATE_

\* Initial state
Init ==
    /\ activePlayers = PLAYERS  \* force all activePlayers and
    /\ activeChairs = CHAIRS    \* activeChairs to be included
    /\ Cardinality(activePlayers) > 1
    /\ Cardinality(activePlayers) = Cardinality(activeChairs) + 1
    /\ occupied = <<>>  \* initially the empty function
    /\ music_playing \in BOOLEAN
    /\ state = "Start"

Walk ==
    /\ state = "Start"
    /\ music_playing
    /\ Cardinality(activePlayers) > 1
    /\ occupied' = <<>>
    /\ state' = "Walking"
    /\ UNCHANGED <<activeChairs, activePlayers, music_playing>>

Sit ==
    /\ state = "Walking"
    /\ ~music_playing
    /\ occupied' \in [activeChairs -> activePlayers]
    \* each chair maps to only one player
    /\ \A c \in activeChairs, p1, p2 \in activePlayers:
        occupied'[c] = p1 /\ occupied'[c] = p2 => p1 = p2
    \* each occupying player sits on one chair
    /\ \A p \in Range(occupied'), c1, c2 \in DOMAIN occupied':
        occupied'[c1] = p /\ occupied'[c2] = p => c1 = c2
    \* there's a player that didn't get to sit down
    \* /\ \E p \in activePlayers: p \notin Range(occupied')
    /\ state' = "Sitting"
    /\ UNCHANGED <<activeChairs, activePlayers, music_playing>>

EliminateLoser ==
    /\ state = "Sitting"
    /\ Cardinality(activePlayers) > 1
    /\ Cardinality(activePlayers) - Cardinality(Range(occupied)) = 1
    /\ activePlayers' = Range(occupied)
    /\ activeChairs' = activeChairs \ {CHOOSE c \in activeChairs : TRUE}
    /\ Cardinality(activeChairs') = Cardinality(activeChairs) - 1
    /\ occupied' = <<>>
    /\ state' = "Start"
    /\ UNCHANGED music_playing

Win ==
    /\ state = "Sitting"
    /\ Cardinality(activePlayers) = 1
    /\ state' = "Won"

ChangeMusicPlaying ==
    /\ music_playing' \in BOOLEAN
    /\ UNCHANGED <<activeChairs, activePlayers, state, occupied>>

\* Safety invariants

OneMorePlayerThanChairs ==
    Cardinality(activePlayers) = Cardinality(activeChairs) + 1

\* Temporal properties

ExistentialLiveness ==
    \E p \in PLAYERS: <> (activePlayers = {p})

FiniteLiveness == <> ENABLED Sit

InfiniteLiveness == <>[](Cardinality(activePlayers) = 1)


Next ==
    \/ Walk
    \/ Sit
    \/ EliminateLoser
    \/ Win
    \/ ChangeMusicPlaying

PlayActions ==
    \/ Walk
    \/ Sit
    \/ EliminateLoser
    \/ Win

Live == SF_vars(PlayActions) /\ WF_vars(ChangeMusicPlaying)
\*Live == TRUE  \* don't assume fairness

Spec == Init /\ [][Next]_vars /\ Live  \* every transition either satisfies the action
                                       \* formula `Next' or leaves the expression
                                       \* `vars' unchanged. In particular, this admits
                                       \* "stuttering transitions" that do not affect
                                       \* `vars'. That is to say,
                                       \* [][Next]_vars == [](Next \/ (vars' = vars)) 

=============================================================================
\* Modification History
\* Last modified Thu Jun 14 23:01:05 EDT 2018 by amin
\* Created Mon May 14 11:12:23 EDT 2018 by amin
