package uw.watform.dash.generator.elements

import uw.watform.dash.dash.Transition
import org.eclipse.xtend.lib.annotations.Accessors
import uw.watform.dash.dash.StateQualNames
import uw.watform.dash.dash.State
import uw.watform.dash.dash.Event
import java.util.HashSet
import java.util.Set
import java.util.ArrayList
import java.util.List

class TransitionGeneratorElement extends GeneratorElement {
    val NAME_PREFIX = "t"

    @Accessors Set<TransitionGeneratorElement> conflicts
    @Accessors Set<TransitionGeneratorElement> transitionsWithHigherPriority

    State arena
    State highestExitedState
    State highestEnteredState

    new (Transition t) {
        this(t, 0)
    }

    new (Transition t, int generatorCounter) {
        super(t, generatorCounter)
        conflicts = new HashSet<TransitionGeneratorElement>
        transitionsWithHigherPriority = new HashSet<TransitionGeneratorElement>
    }

    override getPrefix () {
        NAME_PREFIX
    }

    def State getSourceState() {
        if ((node as Transition).body.source === null) {
            return getParentState
        }
        getState((node as Transition).body.source as StateQualNames)
    }

    def State getTargetState() {
        if ((node as Transition).body.target === null) {
            return getParentState
        }
        getState((node as Transition).body.target as StateQualNames)
    }

    private def getState(StateQualNames states) {
        if (states === null || states.names.isEmpty) {
            return null
        }

        // Return the first element because after the model transformation
        // only one state should be contained
        var state = states.names.get(0)
        if (state instanceof State) {
            return state
        } else {
            return null
        }
    }

    def Event getTriggerEvent() {
        var events = (node as Transition).body.trigger
        if (events === null || events.names.isEmpty) {
            return null
        }

        // Returns the first element because after the transformation only one  
        // Event triggers a transition
        var event = events.names.get(0)
        if (event instanceof Event) {
            return event
        } else {
            return null
        }
    }

    def State getArena() {
        if (this.arena !== null) {
            return this.arena
        }
        // Compute the arena
        var to = getTargetState
        var ancestor = new StateGeneratorElement(getSourceState.parentState ?: sourceState)
        while (ancestor !== null && !ancestor.contains(to)) {
            ancestor = new StateGeneratorElement(ancestor.parentState)
        }
        this.arena = ancestor.node as State
        this.arena
    }

    /**
     * Returns the parent state highest in the hierarchy that is exited by this
     * transition. The retuned state is an immediate child of the arena of the
     * transition
     */
    def State getHighestExitedState() {
        if (highestExitedState === null) {
            highestExitedState =  getHighestConatinerInArena(getSourceState())
        }
        highestExitedState
    }

    /**
     * Returns the parent state highest in the hierarchy that is entered by this
     * transition. The retuned state is an immediate child of the arena of the
     * transition
     */
    def State getHighestEnteredState() {
        if (highestEnteredState === null) {
            highestEnteredState = getHighestConatinerInArena(getTargetState())
        }
        highestEnteredState
    }

    /**
     * Returns the highest state in the arena that contains the provided state
     */
    private def getHighestConatinerInArena(State s) {
        val arena = new StateGeneratorElement(getArena())
        // This happens when a model does not use state hierarchy
        if (arena.node === s) {
            return s
        }
        val children = arena.getSubstates().filter[
            c | Helper.contains(c, s)
        ]
        children.get(0)
    }

    def List<State> getPotentialExitedStates() {
        val exitedStates = new ArrayList<State>()
        var source = getNode.body.source
        if (source instanceof StateQualNames) {
            if ((source).names.get(0) instanceof State) {
                var sourceState = source.names.get(0) as State
                exitedStates.addAll(getExitedStates(sourceState, getHighestExitedState()))
            }
        }
        exitedStates
    }

    def List<State> getExitedStates(State s, State highestExitedState) {
        var exitedStates = new ArrayList<State>()
        exitedStates.add(s)
        var parentState = s.parentState
        if (parentState !== null) {
            if (parentState.isConcurrent) {
                var siblings = Helper.getSiblings(parentState)
                exitedStates.addAll(siblings)
            }
            if (parentState != highestExitedState) {
                exitedStates.addAll(getExitedStates(parentState, highestExitedState))
            }
        }
        exitedStates
    }

    def List<State> getEnteredStates() {
        val enteredStates = new ArrayList<State>()
        val target = getNode.body.target
        if (target instanceof StateQualNames) {
            if ((target).names.get(0) instanceof State) {
                var targetState = target.names.get(0) as State
                enteredStates.addAll(Helper.getEnteredChildrenStates(targetState))
                enteredStates.addAll(getNonDescendantEnteredStates(targetState, getHighestEnteredState()))
            }
        }
        enteredStates
    }

    def List<State> getNonDescendantEnteredStates(State s, State highestEnteredState) {
        var enteredStates = new ArrayList<State>()
        var parentState = s.parentState
        if (parentState !== null && !Helper.contains(parentState, s)) {
            enteredStates.add(parentState)
            if (parentState.isConcurrent) {
                var siblingAncestors = Helper.getSiblings(parentState)
                for (ancestor : siblingAncestors) {
                    enteredStates.addAll(Helper.getEnteredChildrenStates(ancestor))
                }
            }
            enteredStates.addAll(getNonDescendantEnteredStates(parentState, highestEnteredState))
        }
        enteredStates
    }

    override Transition getNode() {
        node as Transition
    }
}
