package uw.watform.dash.scoping

import org.eclipse.xtext.scoping.impl.ImportUriResolver
import org.eclipse.emf.ecore.EObject
import com.google.inject.Inject
import org.eclipse.xtext.resource.FileExtensionProvider
import org.eclipse.emf.common.util.URI
import org.eclipse.emf.ecore.resource.ResourceSet
import com.google.inject.Provider
import java.nio.file.Paths
import java.io.File

class DashImportUriResolver extends ImportUriResolver {

    @Inject
    FileExtensionProvider fileExtensionProvider;

    @Inject
    Provider<ResourceSet> resourceSetProvider;

    public static val UTIL_FOLDER = "util/"

    /**
     * Names of standard util modules shipped with Alloy
     */
    static val UTIL_MODULES = #["boolean", "ctl", "graph", "integer", "natural", "ordering",
                                        "relation", "seqrel", "sequence", "sequniv", "steps", "ternary", "time"]

    /**
     * Resolves the uri of an import object.
     * It returns the resource uri of the import object if the metod is called
     * from Eclipse's editor. Otherwise returns the path to the import file if
     * called in standalone mode.
     * Alloy's standard util modules are provided as a resource inside DASH.
     * Returns null if the import object uri cannot be resolved
     */
    override String apply(EObject from) {
        var apply = super.apply(from)
        if (apply === null) {
            return apply
        }

        var set = resourceSetProvider.get();
        var extensions = fileExtensionProvider.fileExtensions

        for (ext : extensions) {
            try {
                var importUri = URI.createURI(apply + "." +  ext)
                var uri = ""
                if (apply.isUtilModule()) {
                    var classLoader = Thread.currentThread().getContextClassLoader();
                    uri = classLoader.getResource(importUri.toString).toString;
                } else {
                    importUri = importUri.resolve(from.eResource.URI)
                    var importResource = set.getResource(importUri, true);
                    uri = importResource.URI.toString
                }
                return uri
            } catch (RuntimeException e) {
                // The method may have been called in standalone mode
                try {
                    var path = Paths.get(apply + "." +  ext)
                    var filePath = path.toAbsolutePath
                    var file = new File(filePath.toString)
                    if (file.exists && !file.isDirectory) {
                        return filePath.toString
                    }
                } catch (Exception ex) {
                    // File could not be read
                }
            }
        }
        return null
    }

    /**
     * Returns true if the module is a standard module shipped with Alloy
     */
    def isUtilModule(String moduleName) {
        var isUtilModule = false
        if (moduleName.length > 5 && moduleName.substring(0, 5) == UTIL_FOLDER) {
            // Check if the imported module is a standard util module
            for (m : UTIL_MODULES) {
                isUtilModule = isUtilModule || moduleName.equals(UTIL_FOLDER + m)
            }
        }
        isUtilModule
    }
}