package uw.watform.dash.transform

import org.eclipse.emf.ecore.resource.Resource
import uw.watform.dash.dash.State
import java.util.List
import uw.watform.dash.dash.Transition
import org.eclipse.emf.ecore.EObject
import uw.watform.dash.dash.Event
import com.google.inject.Inject
import org.eclipse.xtext.naming.IQualifiedNameProvider
import uw.watform.dash.dash.DashFactory
import uw.watform.dash.dash.DashPackage
import java.util.ArrayList
import uw.watform.dash.dash.TransSet
import uw.watform.dash.dash.Module
import java.util.HashMap
import java.util.Map
import uw.watform.dash.dash.Condition
import uw.watform.dash.dash.Action
import uw.watform.dash.dash.StateItem
import uw.watform.dash.dash.InitDecl
import uw.watform.dash.dash.InvariantDecl
import uw.watform.dash.dash.StateQualNames
import org.eclipse.emf.ecore.util.EcoreUtil
import uw.watform.dash.generator.elements.Helper
import uw.watform.dash.dash.Box
import uw.watform.dash.dash.ElementRef
import uw.watform.dash.dash.TransTemplate
import uw.watform.dash.DashUtil
import uw.watform.dash.dash.EventQualNames
import java.util.HashSet
import java.util.Set
import org.eclipse.emf.ecore.EStructuralFeature.Setting
import uw.watform.dash.generator.elements.TransitionGeneratorElement
import uw.watform.dash.generator.elements.StateGeneratorElement
import uw.watform.dash.dash.Expr
import uw.watform.dash.generator.elements.EventGeneratorElement
import uw.watform.dash.dash.StateVar
import uw.watform.dash.dash.Declaration
import uw.watform.dash.dash.AddOnDecl
import uw.watform.dash.dash.SelectAll
import uw.watform.dash.dash.SelectByName
import uw.watform.dash.dash.SelectByState
import org.eclipse.xtext.EcoreUtil2
import uw.watform.dash.dash.SendEvent

/**
 * Transforms a DASH model to a Core DASH model
 * At the end of the transformation, the patterns of transitions are expanded,
 * the factorings are unfolded and the addons are layered to form the complete
 * definition of transitions.
 *
 * A Core DASH model consists of a state hierarchy with a set of transitions
 * defined at the top level state. Other elements such as events, actions, and
 * conditions are also moved to be defined at the top level state.
 */
class CoreDash implements ITransform {

    @Inject
    extension IQualifiedNameProvider qualifiedNameProvider;

    /**
     * Helper to create generator elements
     */
    @Inject
    extension Helper generatorHelper

    /**
     * Maps {@link State} to {@link TransSet} to be added.
     * During the transformation elements are moved to be declared to a top level state
     */
    var Map<State, List<StateItem>> elementsToAdd = new HashMap<State, List<StateItem>>

    /**
     * Maps {@link State} to {@link TRansSet} to be removed.
     * For example, {@link Addon}s once processed need to be removed from the model
     *
     */
    var Map<State, List<StateItem>> elementsToRemove = new HashMap<State, List<StateItem>>

    /**
     * Maps the transitions generated during their expansion
     */
    var Map<Transition, List<Transition>> generatedtransitionsMap = new HashMap<Transition, List<Transition>>

    // Used to create DASH elements
    DashFactory factory = DashPackage.eINSTANCE.getDashFactory()

    /**
     * Used to give a unique name to every generated {@Link Transition}
     */
    int transition_generation_counter = 1

    /**
     * Transforms a DASH model into a Core DASH model
     */
    override doTransform(Resource resource){
        clear()
        for (m : resource.contents.filter(Module)) {
            m.paragraphs.filter(State).forEach[s | s.transform]
            computeTransitionConflictsAndPriorities(m)
        }
        new Result(resource, generatorHelper.generatorElements)
    }

    private def void clear() {
        transition_generation_counter = 1
        elementsToAdd.clear
        elementsToRemove.clear
        generatedtransitionsMap.clear
        generatorHelper.clear
    }

    private def void computeTransitionConflictsAndPriorities(Module m){
        // Remove elements not needed for generation
        val List<StateItem> removedElements = new ArrayList<StateItem>
        elementsToRemove.values.forEach[l |
            removedElements.addAll(l)
        ]
        removedElements.forEach[ re | generatorHelper.generatorElements.remove(re)]
        val transitions = generatorHelper.generatorElements.values.filter[e |
            e instanceof TransitionGeneratorElement
        ].map[t | t  as TransitionGeneratorElement]

        transitions.forEach[t1 |
            transitions.forEach[t2 |
                if (t1 !== t2) {
                    // Compute conflicts
                    var arena1 = t1.arena?.toGeneratorElement as StateGeneratorElement
                    var arena2 = t2.arena?.toGeneratorElement as StateGeneratorElement
                    if (arena1 !== null && arena2 !==null && arena1.contains(arena2)) {
                        t1.conflicts.add(t2)
                    }
                    // Avoid loops
                    t1.conflicts.add(t1)

                    // Compute priorities
                    var source1 = t1.sourceState?.toGeneratorElement as StateGeneratorElement
                    var source2 = t2.sourceState?.toGeneratorElement as StateGeneratorElement
                    if (source1 !== null && source2 !== null  && source1.strictlyContains(source2)) {
                        t1.transitionsWithHigherPriority.add(t2)
                    }
                }
            ]
        ]
    }

    // Transforms a top level state
    def transform(State s) {
        val List<Transition> transitions = new ArrayList<Transition>()
        mapDescriptiveNames(s)
        // Transitions affected by the State
        s.stateItems.forEach[
            item |
            if (item instanceof TransSet){
                val itemTransitions = item.transform(s, s, s)
                if (itemTransitions !== null) {
                    transitions.addAll(itemTransitions)
                }
            }
        ]

        // Add the transitions at the top Level State and execute queued actions
        s.stateItems.addAll(transitions)
        executeQueuedActions(s)
    }

    // Transforms a substate
    def dispatch List<Transition> transform(State s, TransSet container, State parent, State topLevelState) {
        val List<Transition> transitions = new ArrayList<Transition>()
        mapDescriptiveNames(s)
        // Transitions affected by the State
        s.stateItems.forEach[
            item |
            if (item instanceof TransSet) {
                val itemTransitions = item.transform(s, s, topLevelState)
                if (itemTransitions !== null) {
                    transitions.addAll(itemTransitions)
                }
            }
        ]

        // Remove Transitions
        val transitionsToRemove = s.stateItems.filter[t | t instanceof Transition]
        s.stateItems.removeAll(transitionsToRemove)

        // Add other stateItems (Decl, InitDecl, InvariantDecl) to the top level state
        var otherDeclarations = s.stateItems.reject[item | item instanceof TransSet]
        topLevelState.queueToAdd(otherDeclarations);

        executeQueuedActions(s)

        // Layer Addons
        s.stateItems.filter(AddOnDecl).forEach[a | a.applyAddOn(transitions)]

        return transitions
    }

    private def applyAddOn(AddOnDecl addon, List<Transition> transitions) {
        var applyTo = addon.match
        var List<Transition> modifiedTransitions = new ArrayList<Transition>()
        switch applyTo {
            SelectAll: {modifiedTransitions = transitions}
            SelectByName: {
                modifiedTransitions.addAll(applyTo.names.names)
            }
            SelectByState: {
                val src = applyTo.source
                val tgt = applyTo.target
                if (src !== null && tgt === null) {
                    // TODO implement case "*"
                    if (src instanceof StateQualNames) {
                        modifiedTransitions.addAll(transitions.filter[
                            t | (t.body.source as StateQualNames).names.contains(src.names.get(0))
                        ])
                    }
                } else if (src === null && tgt !== null) {
                    if (tgt instanceof StateQualNames) {
                        modifiedTransitions.addAll(transitions.filter[
                            t | (t.body.target as StateQualNames).names.contains(tgt.names.get(0))
                        ])
                    }
                } else if (src !== null && tgt !== null) {
                    if (src instanceof StateQualNames && tgt instanceof StateQualNames) {
                        modifiedTransitions.addAll(transitions.filter[
                            t | (t.body.source as StateQualNames).names.contains((src as StateQualNames).names.get(0)) &&
                            (t.body.target as StateQualNames).names.contains((tgt as StateQualNames).names.get(0))
                        ])
                    }
                }
            }
        }

        if (addon.addon.action !== null) {
            modifiedTransitions.forEach[ t | t.addAction(addon.addon.action)]
        } else if (addon.addon instanceof SendEvent) {
            modifiedTransitions.forEach[ t | t.addGeneratedEvents((addon.addon as SendEvent).events)]
        }
    }

    private def mapDescriptiveNames(State s) {
        s.stateItems.reject[i |
            i instanceof TransTemplate || i instanceof AddOnDecl
        ].forEach[item |
            var String name = null
            switch item {
                InvariantDecl: name = item.name
                InitDecl: name = item.name
                StateVar: name = ""
                Event, Action, Condition, Transition, State: name = item.name
            }
            item.mapDescriptiveName(name, s)
        ]
        s.mapDescriptiveName(s.name)
    }

    private def executeQueuedActions(State s) {
        // Add other elements
        val elementsToAdd = elementsToAdd.get(s)
        if (elementsToAdd !== null) {
            s.stateItems.addAll(elementsToAdd)
        }

        // Remove other elements
        val elementsToRemove = elementsToRemove.get(s)
        if (elementsToRemove !== null) {
            s.stateItems.removeAll(elementsToRemove)
        }
    }

    def dispatch List<Transition> transform(Event event, TransSet container, State parent, State topLevelState){
        val List<Transition> transitions = new ArrayList<Transition>()

        // Transitins affected by the Event
        event.body.trans.forEach[
            item |
            val itemTransitions = item.transform(event, parent, topLevelState)
            if (itemTransitions !== null) {
                transitions.addAll(itemTransitions)
            }
        ]

        // Add the Trigger Event to the transition
        transitions.forEach[t | t.addTrigger(event)]

        // Remove children except states, they must be transferred to the parent
        // to keep the state hierarchy
        val stateChildren = event.body.trans.filter[e| e instanceof State].map[e| e as StateItem]
        parent.queueToAdd(stateChildren)
        event.body.trans.clear()

        // Add the event declaration to the top level state
        topLevelState.queueToAdd(event)

        return transitions
    }

    def dispatch List<Transition> transform(Condition condition, TransSet container, State parent, State topLevelState){
        val List<Transition> transitions = new ArrayList<Transition>()

        // Transitins affected by the Condition
        condition.body.trans.forEach[
            item |
            val ietmTransitions = item.transform(condition, parent, topLevelState)
            if (ietmTransitions !== null) {
                transitions.addAll(ietmTransitions)
            }
        ]

        // Add the condition forumla to the transitions
        transitions.forEach[t | t.addCondition(condition.formula)]

        // Remove children except states, they must be transferred to the parent
        // to keep the state hierarchy
        val stateChildren = condition.body.trans.filter[e| e instanceof State].map[e| e as StateItem]
        parent.queueToAdd(stateChildren)
        condition.body.trans.clear()

        // Add the event declaration to the top level state
        topLevelState.queueToAdd(condition)

        return transitions
    }

    def  dispatch List<Transition> transform(Action action, TransSet container, State parent, State topLevelState){
        // Add the event declaration to the top level state
        topLevelState.queueToAdd(action)

        // Avoid errors for null return
        return null
    }

    def dispatch List<Transition> transform(Transition t, TransSet container, State parent, State topLevelState) {
        val List<Transition> transitions = new ArrayList<Transition>()
        val subStates = parent.getSubStates
        var List<Transition> toRemove = new ArrayList<Transition>;
        val parentState = t.toGeneratorElement.getParentState

        if (t.body.params !== null) {
            t.expandTemplate
        }

        // complete transition definition
        if (t.body.source === null) {
            t.addSource(parentState)
        }

        if (t.body.target === null) {
            t.addTarget(parentState)
        }

        // Mark events as generated
        if (t.body.genEvents !== null) {
            t.body.genEvents.names?.forEach[e |
                (e.toGeneratorElement as EventGeneratorElement).isGenerated = true
            ]
        }

        // Expand Source
        if (t.body.source !== null && t.body.source.isAll) {
            subStates.forEach[stateElement |
                var copy = generateTransition(t, stateElement, true)
                transitions.add(copy)
                t.mapGeneratedTransitions(copy)
            ]
            parentState.queueToRemove(t)
        } else if (t.body.source instanceof StateQualNames) {
            val sources = (t.body.source as StateQualNames).names
            if (sources.size > 1) {
                sources.forEach[source |
                    if (source instanceof State) {
                        var copy = generateTransition(t, source, true)
                        transitions.add(copy)
                        t.mapGeneratedTransitions(copy)
                    }
                ]
                parentState.queueToRemove(t)
            }
        }

        // Expand Target
        if (t.body.target !== null && t.body.target.isAll){
            var extendedTargetTransitions = new ArrayList<Transition>

            // Make sure the original transition is processed if needed
            if (transitions.empty) {
                transitions.add(t)
            }

            for (transition : transitions){
                for (state : subStates) {
                    var copy = generateTransition(transition, state, false)
                    extendedTargetTransitions.add(copy)
                    toRemove.add(transition)
                }
            }
            t.mapGeneratedTransitions(extendedTargetTransitions)
            transitions.addAll(extendedTargetTransitions)
            parentState.queueToRemove(t)
        } else if (t.body.target instanceof StateQualNames) {
            val targets = (t.body.target as StateQualNames).names
            if (targets.size > 1) {
                val extendedTargetTransitions = new ArrayList<Transition>

                // Make sure the original transition is processed if needed
                if (transitions.empty) {
                    transitions.add(t)
                }

                for (transition : transitions) {
                    for (target: targets){
                        if (target instanceof State) {
                        var copy = generateTransition(transition, target, false)
                        extendedTargetTransitions.add(copy)
                        }
                    }
                    toRemove.add(transition)
                }
                transitions.addAll(extendedTargetTransitions)
                t.mapGeneratedTransitions(extendedTargetTransitions)
                parentState.queueToRemove(t)
            }
        }

        // Remove Transitions
        if (!toRemove.empty) {
            transitions.removeAll(toRemove)
        }

        // No transiions were generated
        if (transitions.empty && !t.body.source.isAll && !t.body.target.isAll) {
            transitions.add(t)
        }

        // Map the names that identifies the original position of the declaration
        transitions.forEach[mapDescriptiveName(name, parentState)]

        transitions
    }

    /** Expand the body of the transition with the contents of the template
     * invocation.
     *
     */
    def private expandTemplate(Transition t) {
        val template = (((t.body.params as Box).left as ElementRef).ref as TransTemplate)
        val Set<Pair<EObject, String>> refNames = new HashSet<Pair<EObject, String>>
        val expandedArguments = new ArrayList<Pair<String, Declaration>>

        template.params.forEach[p | p.names.forEach[ name | expandedArguments.add(name.name -> p)]]
        template.params.forEach[arg |
            var refs = EcoreUtil.UsageCrossReferencer.find(arg, template.body)
            for (ref: refs) {
                var text = DashUtil.getReferenceName(ref.EObject)
                text.split(",").forEach[name |
                    refNames.add(ref.EObject -> name.trim)
                ]
            }
        ]

        // Copy values from template
        if (template.body.source !== null) {
            val element = template.body.source
            val copy = EcoreUtil.copy(element)
            refNames.filter[p | p.key == element].toSet.forEach[
                c | refNames.add(copy -> c.value)
            ]
            t.body.source = copy
        }

        if (template.body.trigger !== null) {
            val element = template.body.trigger
            val copy = EcoreUtil.copy(element)
            refNames.filter[p | p.key == element].toSet.forEach[
                c | refNames.add(copy -> c.value)
            ]
            t.body.trigger = copy
        }

        if (template.body.conditions !== null) {
            t.body.conditions.clear()
            template.body.conditions.forEach[condition |
                val copy = EcoreUtil.copy(condition)
                refNames.filter[p | p.key == condition].toSet.forEach[
                    c | refNames.add(copy -> c.value)
                ]
                t.body.conditions.add(copy)
            ]
        }

        if (template.body.target !== null) {
            val element = template.body.target
            val copy = EcoreUtil.copy(element)
            refNames.filter[p | p.key == element].toSet.forEach[
                c | refNames.add(copy -> c.value)
            ]
            t.body.target = copy
        }

        if (template.body.actions !== null) {
            t.body.actions.clear()
            template.body.actions.forEach[action |
                val copy = EcoreUtil.copy(action)
                refNames.filter[p | p.key == action].toSet.forEach[
                    c | refNames.add(copy -> c.value)
                ]
                t.body.actions.add(copy)
            ]
        }

        if (template.body.genEvents !== null) {
            val element = template.body.genEvents
            val copy = EcoreUtil.copy(element)
            refNames.filter[p | p.key == element].toSet.forEach[
                c | refNames.add(copy -> c.value)
            ]
            t.body.genEvents = copy
        }

        // contains all the references to transition parameters
        var Set<Setting> allReferences = new HashSet<Setting>

        for (var i=0; i< expandedArguments.length; i++) {
            val arg = expandedArguments.get(i)
            var references =  EcoreUtil.UsageCrossReferencer.find(arg.value, t.body)
            allReferences.addAll(references)

            for (ref: references) {
                var matches = refNames.filter[p | p.key == ref.EObject && p.value == arg.key]
                if (matches.length > 0) {
                    // Get real object
                    var actualParam = (t.body.params as Box).args.get(i)
                    var referenceContainer = ref.EObject

                    if (referenceContainer instanceof StateQualNames) {
                        var stateDecl = ((actualParam as ElementRef).ref as State)
                        DashUtil.removeDuplicates(referenceContainer.names)
                        referenceContainer.names.add(stateDecl)
                    }

                    if (referenceContainer instanceof EventQualNames) {
                        var eventeDecl = ((actualParam as ElementRef).ref as Event)
                        DashUtil.removeDuplicates(referenceContainer.names)
                        referenceContainer.names.add(eventeDecl)
                    }
                }
            }
        }

        // Clean references
        for (arg : template.params) {
            for (ref: allReferences) {
                var referenceContainer = ref.EObject
                if (referenceContainer instanceof StateQualNames) {
                    referenceContainer.names.remove(arg)
                } else if (referenceContainer instanceof EventQualNames) {
                     referenceContainer.names.remove(arg)
                }
            }
        }
    }

    def dispatch List<Transition> transform(StateItem t, TransSet container, State parent, State topLevelState) {
        // DO nothing
        null
    }

    /**
     * Generates a {@Link Transition} based on the passed parameter.
     * Creates a copy and then update the source or target according to the
     * isSource parameter
     */
    def private generateTransition(Transition t, State state, boolean isSource) {
        var copy = EcoreUtil.copy(t)
        var names = factory.createStateQualNames();
        names.getNames().add(state);
        names.setAll(false);
        if (isSource) {
            copy.body.source = names
        } else {
            copy.body.target = names
        }
        copy.name = '''«copy.name»_«transition_generation_counter++»'''
        copy
    }

    def private queueToRemove(State container, TransSet element)  {
        queueAction(container, element, true)
    }

    def private queueToAdd(State container, Iterable<StateItem> elements) {
        queueAction(container, elements, false)
    }

    def private queueToAdd(State container, TransSet element)  {
        queueAction(container, element, false)
    }

    def private queueAction(State container, TransSet element, boolean remove) {
        var List<StateItem> elementsList = null
        if (remove) {
            elementsList = elementsToRemove.get(container)
        } else {
            elementsList = elementsToAdd.get(container)
        }

        if (elementsList === null) {
            elementsList = new ArrayList<StateItem>
            if (remove) {
                elementsToRemove.put(container, elementsList)
            } else {
                elementsToAdd.put(container, elementsList)
            }
        }
        elementsList.add(element)
    }

    def private queueAction(State container, Iterable<StateItem> elements, boolean remove) {
        var List<StateItem> elementsList = null
        if (remove) {
            elementsList = elementsToRemove.get(container)
        } else {
            elementsList = elementsToAdd.get(container)
        }

        if (elementsList === null) {
            elementsList = new ArrayList<StateItem>
            if (remove) {
                elementsToRemove.put(container, elementsList)
            } else {
                elementsToAdd.put(container, elementsList)
            }
        }
        elementsList.addAll(elements)
    }

    def private addSource(Transition t, State s) {
        var names = factory.createStateQualNames()
        names.getNames().add(s);
        t.body.setSource(names)
    }

    def private addTarget(Transition t, State s) {
        var names = factory.createStateQualNames()
        names.getNames().add(s);
        t.body.setTarget(names)
    }

    def private addTrigger(Transition t, Event e) {
        if (t.body.trigger === null) {
            var events = factory.createEventQualNames()
            t.body.trigger = events
        }
        t.body.trigger.names.add(e)
    }

    def private addAction(Transition t, Expr action) {
        var copy = EcoreUtil2.copy(action)
        t.body.actions.addAll(copy)
    }

    def private addGeneratedEvents(Transition t, EventQualNames events) {
            if (t.body.genEvents === null) {
            var genEvents = factory.createEventQualNames()
            t.body.genEvents = genEvents
        }
        t.body.genEvents.names.addAll(events.names)
    }

    def private addCondition(Transition t, Expr c) {
        t.body.conditions.add(c)
    }

    def  private Iterable<State> getSubStates(State parent) {
        parent.getChildrenStates
    }

    /**
     * Returns the {@link State} elements that are children of the parameter transSet
     * The children are all state siblings, they are at the same level of the state hierarchy
     *
     * @param transSet  The element for which the state children are returned
     * @return          The list of children states
     */
    def private List<State> getChildrenStates(TransSet transSet) {
        var List<State> stateElements = new ArrayList<State>
        var otherElements = new ArrayList<TransSet>
        var List<TransSet> elements = new ArrayList<TransSet>

        if (transSet instanceof Condition) {
            elements = transSet.body.trans
        } else if (transSet instanceof Event) {
            elements = transSet.body.trans
        } else if (transSet instanceof State) {
            elements = transSet.stateItems.filter[item | item instanceof TransSet].map[i | i as TransSet].toList
        }

        // Add the state elements at this level
        stateElements.addAll(elements.filter[e | e instanceof State].map[e | e as State])
        // Get list of other elements that might contain nested states
        otherElements.addAll(elements.reject[e | e instanceof State])

        // Add StateElements that might be nested deeper
        for(element : otherElements) {
            stateElements.addAll(element.getChildrenStates)
        }
        stateElements
    }

    def private mapDescriptiveName(EObject object,String name) {
        mapDescriptiveName(object, name, object.toGeneratorElement.parentState)
    }

    def private mapDescriptiveName(EObject object, String name, State parent){
        if (parent !== null) {
            var descriptiveName = parent.getFullyQualifiedName
            if (name != "" && name !== null) {
                descriptiveName = descriptiveName.append(name)
            } else if (object instanceof Transition) {
                // Transition names are optional, so we should include the
                // generated name when a name is not given
                descriptiveName = descriptiveName.append(object.toGeneratorElement.shortName)
            }
            object.toGeneratorElement.longName = descriptiveName.toString
        } else {
            if (name != "" && name !== null) {
                object.toGeneratorElement.longName = name
            }
        }
    }

    def private mapGeneratedTransitions(Transition originalTransition, Transition generatedTransition){
        var List<Transition> generatedTransitions = generatedtransitionsMap.get(originalTransition)

        if (generatedTransitions === null) {
            generatedTransitions = new ArrayList<Transition>
        }
        generatedTransitions.add(generatedTransition)
    }

    def private mapGeneratedTransitions(Transition originalTransition, List<Transition> genTransitions) {
       var List<Transition> generatedTransitions = generatedtransitionsMap.get(originalTransition)

        if (generatedTransitions === null) {
            generatedTransitions = new ArrayList<Transition>
        }
        generatedTransitions.addAll(genTransitions)
    }
}
